const std = @import("std");
const rl = @import("raylib");

//Max Vars

pub const MAXEntitys: u32 = 1000;

pub const MAXComponents: u32 = 32;

//Id Types
pub const Entity = usize;

pub const Component = usize;

pub const Signature = std.bit_set.IntegerBitSet(MAXComponents);

//Component array type

pub fn ComponentArray (comptime T: type) type {
    return struct {
        Components: [MAXEntitys]T,
        Entity2Array: std.AutoArrayHashMap(Entity, usize),
        Array2Entity: std.AutoArrayHashMap(usize, Entity),
        sizeM: usize,
        const Self = @This();

        pub fn AddComponent(entity: Entity, comp: T) void {
            var newidx: usize = Self.sizeM;
            Self.Entity2Array[entity] = newidx;
            Self.Entity2Array[newidx] = entity;
            Self.Components[newidx] = comp;
            Self.sizeM += 1;
        }
        pub fn Removedata(entity: Entity) void {
            var remidx: usize = Self.Entity2Array[entity];
            var lastidx: usize = Self.sizeM - 1;
            Self.Components[remidx] = Self.Components[lastidx];

            var lastentity: entity = Self.Array2Entity[lastidx];
            Self.Entity2Array[lastentity] = remidx;
            Self.Array2Entity[remidx] = lastentity;

            Self.Entity2Array[entity] = null;
            Self.Array2Entity[lastentity] = null;

            Self.sizeM -= 1;
        }
        pub fn GetData(entity: Entity) type {
            return Self.Components[Self.Entity2Array[entity]];
        }
        pub fn EntityDestroyed(self: ComponentArray, entity: Entity) void {
            for(self.Array2Entity) |val|{
                if(val == entity){
                    self.Removedata(entity);
                    return;
                }
            }
        }
        pub fn init() Self{
            return .{
                .Components = undefined,
                .Entity2Array = std.AutoArrayHashMap(Entity, usize).init(std.heap.page_allocator),
                .Array2Entity = std.AutoArrayHashMap(usize, Entity).init(std.heap.page_allocator),
                .sizeM = 0
            };
        }
    };
}


//Components

pub const Transform = struct {
    pos: rl.Vector2,
    rot: f16,
};

//Systems

pub const Systems = struct {
    const Self = @This();
    pub fn Gravity() void{

    }
};